



/*
  Unique constructeur de la classe "abstraite" (au sens du CRTP) CDLXAlgorithm.
  Celui-ci fait appel  la fonction "virtuelle" generateSparseMatrix (au sens du CRTP),
  laquelle doit tre redfinie par drivation dans une classe fille de CDLXAlgorithm,
  pour construire la matrice creuse avec laquelle l'algorithme DLX va travailler.

  Argument 1 :
    - type : template<typename> DataT.
    - nom : data.
    - data contient les informations ncessaires pour pouvoir gnrer une matrice creuse.

  Argument 2 (optionnel) :
    - type : unsigned int.
    - nom : maxNbSolutions.
    - maxNbSolutions est le nombre maximal de solutions que l'on souhaite voir calculer
      par l'algorithme DLX. maxNbSolutions est assigne  la donne membre prive m_MaxNbSolutions.
*/
template
  <
  typename DerivedT
  >
template
  <
  typename DataT
  >
CDLXAlgorithm<DerivedT>::CDLXAlgorithm
                                     (
                                     const DataT& data,
                                     const unsigned int maxNbSolutions
                                     ) :
                                        m_MaxNbSolutions
                                                       (
                                                       maxNbSolutions
                                                       )
  {

  generateSparseMatrix
                     (
                     data
                     );

  }



/*
  La fonction cover a plus pour rle d'liminer tous les nuds situs en-dessous de son argument col de type header_t*,
  ainsi que tous les nuds des autres en-tte qui sont lis horizontalement  ceux de col.
  Elle limine ainsi de la matrice creuse toutes les lignes qui font conflit avec col ainsi que col elle-mme.

  Argument :
    - type : header_t* const.
    - nom : col.

  Retour : void.
*/
template
  <
  typename DerivedT
  >
inline void CDLXAlgorithm<DerivedT>::cover
                                         (
                                         header_t* const col
                                         )
  {

// On limine l'en-tte col de la matrice creuse via les 2 instructions suivantes, proposes par Donald Knuth :
  col->right->left=col->left;
  col->left->right=col->right;

// On parcourt tous les noeuds situs en-dessous de col et on les retire de la matrice creuse.
// On limine galement tous les noeuds lis horizontalement  ceux de col en les parcourant vers la droite.
  for(base_node_t* i=col->down;i!=col;i=i->down)
    for(base_node_t* j=i->right;j!=i;j=j->right)
      {

// On limine les noeuds qui font conflit  l'aide des 2 instructions suivantes, proposes par Donald Knuth :
      j->down->up=j->up;
      j->up->down=j->down;
// On n'oublie pas de dcrmenter de 1 le nombre de noeuds d'un en-tte chaque fois qu'on lui en retire un.
      --((matrix_element_t*)j)->header->size;

      }

  }



/*
  La fonction uncover ralise l'opration inverse de cover en rtablissant l'tat prcdent de la matrice.
  Elle traverse un en-tte col de type header_t* vers le haut puis vers la gauche, contrairement  cover
  qui le traverse vers le bas puis vers la droite. uncover rintroduit ainsi les nuds que cover a retirs,
   la place exacte qu'ils occupaient au sein de la matrice creuse.

  Argument :
    - type : header_t* const.
    - nom : col.

  Retour : void.
*/
template
  <
  typename DerivedT
  >
inline void CDLXAlgorithm<DerivedT>::uncover
                                           (
                                           header_t* const col
                                           )
  {

// On introduit les noeuds de bas en haut dans chaque en_tte lesquels sont parcourus
// horizontalement vers la gauche jusqu' celui prcdant col.
  for(base_node_t* i=col->up;i!=col;i=i->up)
    for(base_node_t* j=i->left;j!=i;j=j->left)
      {

// On n'oublie pas d'incrmenter de 1 le nombre de noeuds d'un en-tte chaque fois qu'on insre un noeud dedans.
      ++((matrix_element_t*)j)->header->size;
// On introduit un noeud dans un en-tte de la matrice creuse  l'aide des 2 instructions suivantes, proposes par Donald Knuth :
      j->down->up=j;
      j->up->down=j;

      }

// On introduit l'en-tte col dans la matrice creuse  l'aide des 2 instructions suivantes, proposes par Donald Knuth :
  col->right->left=col;
  col->left->right=col;

  }



/*
  La fonction chooseColumn, permet de slectionner, l'en-tte de colonne (de type header_t)
  qui liste le moins de nuds (c'est--dire la colonne comptant le moins de 1)
  pour minimiser le nombre de branchements dans l'algorithme DLX.

  Argument : aucun.

  Retour :
    - type :header_t*.
    - nom : col.
    - col liste le moins de noeuds (donne membre size la plus petite).
*/
template
  <
  typename DerivedT
  >
inline typename CDLXAlgorithm<DerivedT>::header_t* CDLXAlgorithm<DerivedT>::chooseColumn()
  {

  // On dclare un entier s initialis  la valeur maximale des unsigned int.
  unsigned int s=std::numeric_limits<unsigned int>::max();
  // On dclaration l'en-tte col qui sera retourn.
  header_t* col=0;
  // On recherche  droite de la racine H l'en-tte listant le moins de noeuds.
  for(header_t* j=(header_t*)m_H->right;j!=m_H;j=(header_t*)j->right)
    if(j->size<s) // On vrifie si le nombre de nneuds list par l'en-tte courant j est plus petit que s.
      s=(col=j)->size; // Si c'est le cas, col est assign  j et s au champs size de j.
  // On retourne l'en-tte col.
  return
       (
       col
       );

  }



/*
  La fonction search implmente l'algorithme DLX de Donald Knuth et qui est, en fait,
  une "spcialisation" de l'algorithme X (invent aussi par Donald Knuth).
  Cet algorithm permet de trouver toutes les solutions d'un problme de couverture exacte.
  Il est, de plus, de force brute, rcursif et de profondeur d'abord.

  Argument : aucun.

  Retour : void.
*/
template
  <
  typename DerivedT
  >
inline void CDLXAlgorithm<DerivedT>::search()
  {

//  Si on a dj atteint le nombre maximal de solutions
//  que l'on souhaitait voir calcules par l'algorithme DLX,
//  on sort de la fonction search.
  if(m_Solutions.size()>=m_MaxNbSolutions)
    return;

//  On teste la condition d'arrt de rcursion de l'algorithme DLX.
//  S'il n'y a plus d'en-tte ) tester, on enregistre la solution et
// on sort de la fonction search.
  if(m_H->right==m_H)
    {
    m_Solutions.push_back
                        (
                        m_O
                        );
    return;
    }

//  On choisit l'en-tte qui liste le moins de noeuds, en
//  appelant la fonction chooseColumn. Cela permet de minimiser
//  la profondeur de rcursion.
  header_t* col=chooseColumn();

//  Si on trouve un en-tte dant la liste de noeuds est vide,
//  (champs size==0), alors cette branche de rcursion de
//  l'algorithme DLX n'a pas de solution; on sort alors de la
//  fonction search.
  if((!col)||(!col->size))
    return;
  cover
      (
      col
      );

//  On supprime de la matrice la colonne qui fait conflit.
  for(base_node_t* r=col->down;r!=col;r=r->down)
    {

//  On ajoute r  la solution partielle O.
    m_O.push_back
                (
                ((matrix_element_t*)r)->row
                );
//  On rduit la matrice en enlevant les lignes qui font conflit avec la colonne.
    for(base_node_t* j=r->right;j!=r;j=j->right)
      cover
          (
          ((matrix_element_t*)j)->header
          );

//  On appelle search rcursivement. C'est la partie non dterministe de
//  l'algorithme DLX. En effet, chaque sous-fonction search hrite de la mme
//  matrice mais la rduit selon la valeur de r.
    search();

//  On enlve r de la solution partielle O;
    m_O.pop_back();

//  On rtablit les lignes de la matrice par appel  uncover.
    for(base_node_t* j=r->left;j!=r;j=j->left)
      uncover
            (
            ((matrix_element_t*)j)->header
            );

    }

//  On rtablit la colonne liste par l'en-tte col
//  en appelant uncover
  uncover
        (
        col
        );

  }



/*
  La fonction membre publique getIthSolution retourne une solution calcule par search dans un conteneur associatif de type
  std::map<char,std::vector<unsigned int> >. les clefs sont de type char et correspondent aux pices A, B, L, P, T et Z
  du jeu SOMA. Les valeurs, de type std::vector<unsigned int>, sont les listes des cubes qu'occupent la pice en question
  dans la structure rsolue. Ainsi, par exemple, si m est une instance d'un tel conteneur, m[V] sera une liste de 3
  m[A] aura 4 lments.

  Argument : aucun.

  Retour :
    - type : const std::map<char,std::vector<unsigned int> >.
    - nom : ithSolution.
*/
template
  <
  typename DerivedT
  >
inline const std::map<char,std::vector<unsigned int> > CDLXAlgorithm<DerivedT>::getIthSolution
                                                                                             (
                                                                                             const unsigned int solutionIndex
                                                                                             ) const
  {

//  On lance une exception de type std::out_of_range si l'indice de la solution que l'on souhaite
//  rcuprer dpasse le nombre de solutions qu'a trouves l'algorithme DLX.
  if(((int)solutionIndex<0)||(solutionIndex>m_Solutions.size()))
    throw
        (
        std::out_of_range
                        (
                        "Solution index not in range"
                        )
        );

  std::map<char,std::vector<unsigned int> > ithSolution;

  const unsigned int nbRowsInSolution=m_Solutions[solutionIndex].size();
  for(unsigned int i=0;i<nbRowsInSolution;++i)
    {

//  Par construction de la structure toroidale (la matrice creuse), on sait que le premier noeud
//  d'une ligne correspond  un 1 dans la colonne de l'une des pices du jeu SOMA. Ainsi, l'on
//  est sr que le champ name de l'en-tte dudit noeud contient le nom de la pice en question (A, B, L, P, T, V ou Z).
    const char pieceName=m_Rows[m_Solutions[solutionIndex][i]][0]->header->name;
    const unsigned int nbOnesInRow=m_Rows[m_Solutions[solutionIndex][i]].size();

//  Les autres noeuds de m_Rows[m_Solutions[solutionIndex][i]][j] sont, quant,  eux, des positions (pas forcments tous
//  sur la mme ligne) des cubes de la pice dans sa structure. On doit donc retenir le champs column de l'en-tte de ces noeuds.
    for(unsigned int j=1;j<nbOnesInRow;++j)
      ithSolution[pieceName].push_back
                                     (
                                     m_Rows[m_Solutions[solutionIndex][i]][j]->header->column
                                     );

    }

//  On retourne la solution.
  return
       (
       ithSolution
       );

  }
